# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'synthesis_view.ui'
#
# Created by: PyQt5 UI code generator 5.14.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_synthesis_view(object):
    def setupUi(self, synthesis_view):
        synthesis_view.setObjectName("synthesis_view")
        synthesis_view.resize(523, 622)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(synthesis_view)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea = QtWidgets.QScrollArea(synthesis_view)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 558, 584))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.main_title = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.main_title.sizePolicy().hasHeightForWidth())
        self.main_title.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.main_title.setFont(font)
        self.main_title.setScaledContents(False)
        self.main_title.setAlignment(QtCore.Qt.AlignCenter)
        self.main_title.setObjectName("main_title")
        self.verticalLayout_3.addWidget(self.main_title)
        self.configurationBox = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.configurationBox.setObjectName("configurationBox")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.configurationBox)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.configurationVLayout = QtWidgets.QVBoxLayout()
        self.configurationVLayout.setObjectName("configurationVLayout")
        self.traceLayout = QtWidgets.QHBoxLayout()
        self.traceLayout.setObjectName("traceLayout")
        self.trace_label = QtWidgets.QLabel(self.configurationBox)
        self.trace_label.setObjectName("trace_label")
        self.traceLayout.addWidget(self.trace_label)
        self.trace_type_box = QtWidgets.QComboBox(self.configurationBox)
        self.trace_type_box.setObjectName("trace_type_box")
        self.traceLayout.addWidget(self.trace_type_box)
        self.trace_line = QtWidgets.QLineEdit(self.configurationBox)
        self.trace_line.setObjectName("trace_line")
        self.traceLayout.addWidget(self.trace_line)
        self.configurationVLayout.addLayout(self.traceLayout)
        self.targetLayout = QtWidgets.QHBoxLayout()
        self.targetLayout.setObjectName("targetLayout")
        self.from_label = QtWidgets.QLabel(self.configurationBox)
        self.from_label.setEnabled(False)
        self.from_label.setObjectName("from_label")
        self.targetLayout.addWidget(self.from_label)
        self.from_line = QtWidgets.QLineEdit(self.configurationBox)
        self.from_line.setEnabled(False)
        self.from_line.setPlaceholderText("")
        self.from_line.setObjectName("from_line")
        self.targetLayout.addWidget(self.from_line)
        self.to_label = QtWidgets.QLabel(self.configurationBox)
        self.to_label.setEnabled(False)
        self.to_label.setObjectName("to_label")
        self.targetLayout.addWidget(self.to_label)
        self.to_line = QtWidgets.QLineEdit(self.configurationBox)
        self.to_line.setEnabled(False)
        self.to_line.setObjectName("to_line")
        self.targetLayout.addWidget(self.to_line)
        self.target_label = QtWidgets.QLabel(self.configurationBox)
        self.target_label.setEnabled(False)
        self.target_label.setObjectName("target_label")
        self.targetLayout.addWidget(self.target_label)
        self.target_box = QtWidgets.QComboBox(self.configurationBox)
        self.target_box.setEnabled(False)
        self.target_box.setObjectName("target_box")
        self.targetLayout.addWidget(self.target_box)
        self.register_box = QtWidgets.QComboBox(self.configurationBox)
        self.register_box.setEnabled(False)
        self.register_box.setObjectName("register_box")
        self.targetLayout.addWidget(self.register_box)
        self.mem_line = QtWidgets.QLineEdit(self.configurationBox)
        self.mem_line.setEnabled(False)
        self.mem_line.setObjectName("mem_line")
        self.targetLayout.addWidget(self.mem_line)
        self.operand_label = QtWidgets.QLabel(self.configurationBox)
        self.operand_label.setEnabled(False)
        self.operand_label.setObjectName("operand_label")
        self.targetLayout.addWidget(self.operand_label)
        self.configurationVLayout.addLayout(self.targetLayout)
        self.tableLayout = QtWidgets.QHBoxLayout()
        self.tableLayout.setObjectName("tableLayout")
        self.table_label = QtWidgets.QLabel(self.configurationBox)
        self.table_label.setEnabled(False)
        self.table_label.setObjectName("table_label")
        self.tableLayout.addWidget(self.table_label)
        self.table_type_box = QtWidgets.QComboBox(self.configurationBox)
        self.table_type_box.setEnabled(False)
        self.table_type_box.setObjectName("table_type_box")
        self.tableLayout.addWidget(self.table_type_box)
        self.table_line = QtWidgets.QLineEdit(self.configurationBox)
        self.table_line.setEnabled(False)
        self.table_line.setObjectName("table_line")
        self.tableLayout.addWidget(self.table_line)
        self.configurationVLayout.addLayout(self.tableLayout)
        self.algorithmLayout = QtWidgets.QHBoxLayout()
        self.algorithmLayout.setObjectName("algorithmLayout")
        self.algorithm_label = QtWidgets.QLabel(self.configurationBox)
        self.algorithm_label.setEnabled(False)
        self.algorithm_label.setObjectName("algorithm_label")
        self.algorithmLayout.addWidget(self.algorithm_label)
        self.algorithm_box = QtWidgets.QComboBox(self.configurationBox)
        self.algorithm_box.setEnabled(False)
        self.algorithm_box.setObjectName("algorithm_box")
        self.algorithmLayout.addWidget(self.algorithm_box)
        self.algorithm_type_label = QtWidgets.QLabel(self.configurationBox)
        self.algorithm_type_label.setEnabled(False)
        self.algorithm_type_label.setObjectName("algorithm_type_label")
        self.algorithmLayout.addWidget(self.algorithm_type_label)
        self.algorithm_type_box = QtWidgets.QComboBox(self.configurationBox)
        self.algorithm_type_box.setEnabled(False)
        self.algorithm_type_box.setObjectName("algorithm_type_box")
        self.algorithmLayout.addWidget(self.algorithm_type_box)
        self.qtrace_sym_type_box = QtWidgets.QComboBox(self.configurationBox)
        self.qtrace_sym_type_box.setEnabled(False)
        self.qtrace_sym_type_box.setObjectName("qtrace_sym_type_box")
        self.algorithmLayout.addWidget(self.qtrace_sym_type_box)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.algorithmLayout.addItem(spacerItem)
        self.configurationVLayout.addLayout(self.algorithmLayout)
        self.experimentalLayout = QtWidgets.QHBoxLayout()
        self.experimentalLayout.setObjectName("experimentalLayout")
        self.label_9 = QtWidgets.QLabel(self.configurationBox)
        self.label_9.setObjectName("label_9")
        self.experimentalLayout.addWidget(self.label_9)
        self.verticalLayout_9 = QtWidgets.QVBoxLayout()
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.learning_checkbox = QtWidgets.QCheckBox(self.configurationBox)
        self.learning_checkbox.setObjectName("learning_checkbox")
        self.verticalLayout_9.addWidget(self.learning_checkbox)
        self.check_sem_checkbox = QtWidgets.QCheckBox(self.configurationBox)
        self.check_sem_checkbox.setObjectName("check_sem_checkbox")
        self.verticalLayout_9.addWidget(self.check_sem_checkbox)
        self.experimentalLayout.addLayout(self.verticalLayout_9)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.experimentalLayout.addItem(spacerItem1)
        self.configurationVLayout.addLayout(self.experimentalLayout)
        self.verticalLayout_6.addLayout(self.configurationVLayout)
        self.verticalLayout_3.addWidget(self.configurationBox)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.tritonLayout = QtWidgets.QVBoxLayout()
        self.tritonLayout.setObjectName("tritonLayout")
        self.run_triton_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.run_triton_button.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.run_triton_button.sizePolicy().hasHeightForWidth())
        self.run_triton_button.setSizePolicy(sizePolicy)
        self.run_triton_button.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.run_triton_button.setFlat(False)
        self.run_triton_button.setObjectName("run_triton_button")
        self.tritonLayout.addWidget(self.run_triton_button, 0, QtCore.Qt.AlignHCenter)
        self.triton_textarea = QtWidgets.QTextBrowser(self.scrollAreaWidgetContents)
        self.triton_textarea.setEnabled(False)
        self.triton_textarea.setObjectName("triton_textarea")
        self.tritonLayout.addWidget(self.triton_textarea)
        self.triton_button_group = QtWidgets.QHBoxLayout()
        self.triton_button_group.setObjectName("triton_button_group")
        self.show_deps_triton_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.show_deps_triton_button.setEnabled(False)
        self.show_deps_triton_button.setObjectName("show_deps_triton_button")
        self.triton_button_group.addWidget(self.show_deps_triton_button)
        self.line_2 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.triton_button_group.addWidget(self.line_2)
        self.show_ast_triton_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.show_ast_triton_button.setEnabled(False)
        self.show_ast_triton_button.setObjectName("show_ast_triton_button")
        self.triton_button_group.addWidget(self.show_ast_triton_button)
        self.tritonLayout.addLayout(self.triton_button_group)
        self.horizontalLayout_2.addLayout(self.tritonLayout)
        self.line = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout_2.addWidget(self.line)
        self.synthesisLayout = QtWidgets.QVBoxLayout()
        self.synthesisLayout.setObjectName("synthesisLayout")
        self.run_synthesis_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.run_synthesis_button.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.run_synthesis_button.sizePolicy().hasHeightForWidth())
        self.run_synthesis_button.setSizePolicy(sizePolicy)
        self.run_synthesis_button.setObjectName("run_synthesis_button")
        self.synthesisLayout.addWidget(self.run_synthesis_button, 0, QtCore.Qt.AlignHCenter)
        self.synthesis_textarea = QtWidgets.QTextBrowser(self.scrollAreaWidgetContents)
        self.synthesis_textarea.setEnabled(False)
        self.synthesis_textarea.setObjectName("synthesis_textarea")
        self.synthesisLayout.addWidget(self.synthesis_textarea)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.show_ast_synthesis_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.show_ast_synthesis_button.setEnabled(False)
        self.show_ast_synthesis_button.setObjectName("show_ast_synthesis_button")
        self.horizontalLayout_4.addWidget(self.show_ast_synthesis_button)
        self.line_3 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.line_3.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.horizontalLayout_4.addWidget(self.line_3)
        self.reassemble_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.reassemble_button.setEnabled(False)
        self.reassemble_button.setObjectName("reassemble_button")
        self.horizontalLayout_4.addWidget(self.reassemble_button)
        self.synthesisLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_2.addLayout(self.synthesisLayout)
        self.verticalLayout_3.addLayout(self.horizontalLayout_2)
        self.verticalLayout_4.addLayout(self.verticalLayout_3)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(synthesis_view)
        QtCore.QMetaObject.connectSlotsByName(synthesis_view)

    def retranslateUi(self, synthesis_view):
        _translate = QtCore.QCoreApplication.translate
        synthesis_view.setWindowTitle(_translate("synthesis_view", "QSynthesis"))
        self.main_title.setText(_translate("synthesis_view", "QSynthesis"))
        self.configurationBox.setTitle(_translate("synthesis_view", "Synthesis configuration"))
        self.trace_label.setText(_translate("synthesis_view", "Trace:"))
        self.from_label.setText(_translate("synthesis_view", "From:"))
        self.to_label.setText(_translate("synthesis_view", "To:"))
        self.target_label.setText(_translate("synthesis_view", "Target:"))
        self.operand_label.setText(_translate("synthesis_view", "TextLabel"))
        self.table_label.setText(_translate("synthesis_view", "Table:"))
        self.algorithm_label.setText(_translate("synthesis_view", "Algorithm:"))
        self.algorithm_type_label.setText(_translate("synthesis_view", "Type:"))
        self.label_9.setText(_translate("synthesis_view", "Experimental settings:"))
        self.learning_checkbox.setText(_translate("synthesis_view", "Learning new expressions"))
        self.check_sem_checkbox.setText(_translate("synthesis_view", "Check semantic expressions"))
        self.run_triton_button.setText(_translate("synthesis_view", "Run Triton"))
        self.show_deps_triton_button.setText(_translate("synthesis_view", "Show deps"))
        self.show_ast_triton_button.setText(_translate("synthesis_view", "Show AST"))
        self.run_synthesis_button.setText(_translate("synthesis_view", "Run Synthesis"))
        self.show_ast_synthesis_button.setText(_translate("synthesis_view", "Show AST"))
        self.reassemble_button.setText(_translate("synthesis_view", "Reassemble"))
